from PyQt5.QtCore import QObject, pyqtSignal

from kamzik3.snippets.snippetsWidgets import PyQtSignalEmitter


class TestPyQtSignalEmitter:
    class A(QObject):
        signalA = pyqtSignal()

    a1 = A()
    a2 = A()

    e1a1 = PyQtSignalEmitter(a1)
    e2a1 = PyQtSignalEmitter(a1)
    e3a2 = PyQtSignalEmitter(a1)

    def test_emitters_on_same_signal_compare_equal(self):
        assert PyQtSignalEmitter(self.a1.signalA) == PyQtSignalEmitter(self.a1.signalA)

    def test_emitters_on_different_signals_compare_unequal(self):
        assert PyQtSignalEmitter(self.a1.signalA) != PyQtSignalEmitter(self.a2.signalA)

    def test_emitters_on_same_signal_has_same_hash(self):
        assert hash(PyQtSignalEmitter(self.a1.signalA)) == hash(
            PyQtSignalEmitter(self.a1.signalA)
        )

    def test_emitters_on_different_signals_have_different_hashes(self):
        assert hash(PyQtSignalEmitter(self.a1.signalA)) != hash(
            PyQtSignalEmitter(self.a2.signalA)
        )

    def test_emitters_on_same_signal_overwrite_dict_key(self):
        d = {}
        d[PyQtSignalEmitter(self.a1.signalA)] = 1
        d[PyQtSignalEmitter(self.a1.signalA)] = 3
        assert len(d) == 1

    def test_emitters_on_different_signals_dont_overwrite_dict_key(self):
        d = {}
        d[PyQtSignalEmitter(self.a1.signalA)] = 1
        d[PyQtSignalEmitter(self.a2.signalA)] = 3
        assert len(d) == 2

    def test_remove_from_dict(self):
        d = {
            PyQtSignalEmitter(self.a1.signalA): 1,
            PyQtSignalEmitter(self.a2.signalA): 3,
        }
        del d[PyQtSignalEmitter(self.a1.signalA)]
        assert len(d) == 1
        del d[PyQtSignalEmitter(self.a2.signalA)]
        assert len(d) == 0
